//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Most of the code from this file comes from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//			(file: agg_svg_parser.cpp)
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include <search.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>
#include "expat.h"
#include "agg_basics.h"
#include "svg_exception.h"
#include "svg_parser.h"
#include "svg_element.h"
#include "svg_group.h"
#include "svg_path.h"
#include "svg_dash.h"
#include "svg_font_win32.h"
#include "svg_text.h"
#include "svg_tspan.h"
#include "svg_text_path.h"
#include "svg_defs.h"
#include "svg_tref.h"
#include "svg_stop.h"
#include "svg_linear_gradient.h"
#include "svg_radial_gradient.h"
#include "svg_pattern.h"
#include "svg_use.h"
#include "svg_filter.h"
#include "svg_fe_gaussian_blur.h"
#include "svg_clip_path.h"
#include "svg_mask.h"
#include "svg_poly.h"
#include "svg_line.h"
#include "svg_ellipse.h"
#include "svg_rect.h"

#ifdef TEX_EXTENSIONS
//TEX_EXTENSIONS: incorporates additional elements, which handles 
//special xml files. 
//	tex:scr	-		includes content of an SVG file as a group in the
//					final picture. Enables composition of many SVG files
//					into a single drawing.	
//	tex:override  - overwrite a style, or filter, or mask, or clip path settings 
//					for an element in an svg file (to control the visibility of 
//					subsequent layers, for example) .
#include "tex_src.h"
#include "tex_override.h"
#endif //TEX_EXTENSIONS

#include "svg_svg.h" //this include always place at the end
					 //becasue after it you cannot use the namespace 
					 //qualifier (svg::)
namespace svg
{
	//////////////////////////////////////////////////////////////////////
	// local helper functions and data (handling colors, etc.)
	//////////////////////////////////////////////////////////////////////
    struct named_color
    {
        char  name[22];
		agg::int8u r, g, b, a;
    };

    static named_color colors[] = 
    {
        { "aliceblue",240,248,255, 255 },
        { "antiquewhite",250,235,215, 255 },
        { "aqua",0,255,255, 255 },
        { "aquamarine",127,255,212, 255 },
        { "azure",240,255,255, 255 },
        { "beige",245,245,220, 255 },
        { "bisque",255,228,196, 255 },
        { "black",0,0,0, 255 },
        { "blanchedalmond",255,235,205, 255 },
        { "blue",0,0,255, 255 },
        { "blueviolet",138,43,226, 255 },
        { "brown",165,42,42, 255 },
        { "burlywood",222,184,135, 255 },
        { "cadetblue",95,158,160, 255 },
        { "chartreuse",127,255,0, 255 },
        { "chocolate",210,105,30, 255 },
        { "coral",255,127,80, 255 },
        { "cornflowerblue",100,149,237, 255 },
        { "cornsilk",255,248,220, 255 },
        { "crimson",220,20,60, 255 },
        { "cyan",0,255,255, 255 },
        { "darkblue",0,0,139, 255 },
        { "darkcyan",0,139,139, 255 },
        { "darkgoldenrod",184,134,11, 255 },
        { "darkgray",169,169,169, 255 },
        { "darkgreen",0,100,0, 255 },
        { "darkgrey",169,169,169, 255 },
        { "darkkhaki",189,183,107, 255 },
        { "darkmagenta",139,0,139, 255 },
        { "darkolivegreen",85,107,47, 255 },
        { "darkorange",255,140,0, 255 },
        { "darkorchid",153,50,204, 255 },
        { "darkred",139,0,0, 255 },
        { "darksalmon",233,150,122, 255 },
        { "darkseagreen",143,188,143, 255 },
        { "darkslateblue",72,61,139, 255 },
        { "darkslategray",47,79,79, 255 },
        { "darkslategrey",47,79,79, 255 },
        { "darkturquoise",0,206,209, 255 },
        { "darkviolet",148,0,211, 255 },
        { "deeppink",255,20,147, 255 },
        { "deepskyblue",0,191,255, 255 },
        { "dimgray",105,105,105, 255 },
        { "dimgrey",105,105,105, 255 },
        { "dodgerblue",30,144,255, 255 },
        { "firebrick",178,34,34, 255 },
        { "floralwhite",255,250,240, 255 },
        { "forestgreen",34,139,34, 255 },
        { "fuchsia",255,0,255, 255 },
        { "gainsboro",220,220,220, 255 },
        { "ghostwhite",248,248,255, 255 },
        { "gold",255,215,0, 255 },
        { "goldenrod",218,165,32, 255 },
        { "gray",128,128,128, 255 },
        { "green",0,128,0, 255 },
        { "greenyellow",173,255,47, 255 },
        { "grey",128,128,128, 255 },
        { "honeydew",240,255,240, 255 },
        { "hotpink",255,105,180, 255 },
        { "indianred",205,92,92, 255 },
        { "indigo",75,0,130, 255 },
        { "ivory",255,255,240, 255 },
        { "khaki",240,230,140, 255 },
        { "lavender",230,230,250, 255 },
        { "lavenderblush",255,240,245, 255 },
        { "lawngreen",124,252,0, 255 },
        { "lemonchiffon",255,250,205, 255 },
        { "lightblue",173,216,230, 255 },
        { "lightcoral",240,128,128, 255 },
        { "lightcyan",224,255,255, 255 },
        { "lightgoldenrodyellow",250,250,210, 255 },
        { "lightgray",211,211,211, 255 },
        { "lightgreen",144,238,144, 255 },
        { "lightgrey",211,211,211, 255 },
        { "lightpink",255,182,193, 255 },
        { "lightsalmon",255,160,122, 255 },
        { "lightseagreen",32,178,170, 255 },
        { "lightskyblue",135,206,250, 255 },
        { "lightslategray",119,136,153, 255 },
        { "lightslategrey",119,136,153, 255 },
        { "lightsteelblue",176,196,222, 255 },
        { "lightyellow",255,255,224, 255 },
        { "lime",0,255,0, 255 },
        { "limegreen",50,205,50, 255 },
        { "linen",250,240,230, 255 },
        { "magenta",255,0,255, 255 },
        { "maroon",128,0,0, 255 },
        { "mediumaquamarine",102,205,170, 255 },
        { "mediumblue",0,0,205, 255 },
        { "mediumorchid",186,85,211, 255 },
        { "mediumpurple",147,112,219, 255 },
        { "mediumseagreen",60,179,113, 255 },
        { "mediumslateblue",123,104,238, 255 },
        { "mediumspringgreen",0,250,154, 255 },
        { "mediumturquoise",72,209,204, 255 },
        { "mediumvioletred",199,21,133, 255 },
        { "midnightblue",25,25,112, 255 },
        { "mintcream",245,255,250, 255 },
        { "mistyrose",255,228,225, 255 },
        { "moccasin",255,228,181, 255 },
        { "navajowhite",255,222,173, 255 },
        { "navy",0,0,128, 255 },
        { "oldlace",253,245,230, 255 },
        { "olive",128,128,0, 255 },
        { "olivedrab",107,142,35, 255 },
        { "orange",255,165,0, 255 },
        { "orangered",255,69,0, 255 },
        { "orchid",218,112,214, 255 },
        { "palegoldenrod",238,232,170, 255 },
        { "palegreen",152,251,152, 255 },
        { "paleturquoise",175,238,238, 255 },
        { "palevioletred",219,112,147, 255 },
        { "papayawhip",255,239,213, 255 },
        { "peachpuff",255,218,185, 255 },
        { "peru",205,133,63, 255 },
        { "pink",255,192,203, 255 },
        { "plum",221,160,221, 255 },
        { "powderblue",176,224,230, 255 },
        { "purple",128,0,128, 255 },
        { "red",255,0,0, 255 },
        { "rosybrown",188,143,143, 255 },
        { "royalblue",65,105,225, 255 },
        { "saddlebrown",139,69,19, 255 },
        { "salmon",250,128,114, 255 },
        { "sandybrown",244,164,96, 255 },
        { "seagreen",46,139,87, 255 },
        { "seashell",255,245,238, 255 },
        { "sienna",160,82,45, 255 },
        { "silver",192,192,192, 255 },
        { "skyblue",135,206,235, 255 },
        { "slateblue",106,90,205, 255 },
        { "slategray",112,128,144, 255 },
        { "slategrey",112,128,144, 255 },
        { "snow",255,250,250, 255 },
        { "springgreen",0,255,127, 255 },
        { "steelblue",70,130,180, 255 },
        { "tan",210,180,140, 255 },
        { "teal",0,128,128, 255 },
        { "thistle",216,191,216, 255 },
        { "tomato",255,99,71, 255 },
        { "turquoise",64,224,208, 255 },
        { "violet",238,130,238, 255 },
        { "wheat",245,222,179, 255 },
        { "white",255,255,255, 255 },
        { "whitesmoke",245,245,245, 255 },
        { "yellow",255,255,0, 255 },
        { "yellowgreen",154,205,50, 255 },
        { "zzzzzzzzzzz",0,0,0, 0 }
    }; 

    //-------------------------------------------------------------
    static int cmp_color(const void* p1, const void* p2)
    {
        return strcmp(((named_color*)p1)->name, ((named_color*)p2)->name);
    }

    //-------------------------------------------------------------
	static agg::rgba8 parse_color(const char* str)
    {
        while(*str == ' ') ++str;
        unsigned c = 0;
        if(*str == '#')
        {
            sscanf(str + 1, "%x", &c);
            return agg::rgb8_packed(c);
        }
        else if(strncmp(str, "rgb(", 4) == 0)
        {
            unsigned r, g, b;
            sscanf(str + 4, "%d,%d,%d", &r, &g, &b);
            return agg::rgba8(r, g, b);
        }
        else
        {
            named_color c;
            unsigned len = strlen(str);
            if(len > sizeof(c.name) - 1)
            {
                throw exception("parse_color: Invalid color name '%s'", str);
            }
            strcpy(c.name, str);
            const void* p = bsearch(&c, 
                                    colors, 
                                    sizeof(colors) / sizeof(colors[0]), 
                                    sizeof(colors[0]), 
                                    cmp_color);
            if(p == 0)
            {
                throw exception("parse_color: Invalid color name '%s'", str);
            }
            const named_color* pc = (const named_color*)p;
            return agg::rgba8(pc->r, pc->g, pc->b, pc->a);
        }
    }

    //-------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////
//
// Implementation of svg::parser
//
///////////////////////////////////////////////////////////////////////////
	//Arguments:
	//	paths:	path storage, that will be filled with element's paths
	//	attrs:	attribute storage, that will be filled with element's attributes
	//	elems:	element storage, that will be filled with the SVG graph elements
	//			(instances are created on the heap, using new operator)
	parser::parser(path_storage &paths, attr_storage &attrs, element_storage &elems):
		m_paths(paths), m_attrs(attrs), m_elems(elems)
	{
			m_element = NULL;
			m_level = 0;
	}
	
    //-------------------------------------------------------------
	//Destructor
	parser::~parser() 
	{}
    //-------------------------------------------------------------

	//main method - parse the given file name
	void parser::parse(const char *fname)
    {
		char buf[1024], msg[512];
		
		setlocale( LC_ALL, "English" ); //all decimal points are U.S. English!
	    
		XML_Parser p = XML_ParserCreate(NULL);
	    if(p == 0) 
	    {
		    throw exception("Couldn't allocate memory for parser");
	    }

        XML_SetUserData(p, this);
	    XML_SetElementHandler(p, start_element, end_element);
	    XML_SetCharacterDataHandler(p, content);

        FILE* fd = fopen(fname, "r");
        if(fd == 0)
        {
            sprintf(msg, "Couldn't open file %s", fname);
		    throw exception(msg);
        }

        bool done = false;
        do
        {
            size_t len = fread(buf, 1, sizeof(buf), fd);
            done = len < sizeof(buf);
            if(!XML_Parse(p, buf, len, done))
            {
                sprintf(msg,
                    "%s at line %d\n",
                    XML_ErrorString(XML_GetErrorCode(p)),
                    XML_GetCurrentLineNumber(p));
                throw exception(msg);
            }
        }
        while(!done);
        fclose(fd);
        XML_ParserFree(p);
    }//parser::parse

    //-------------------------------------------------------------
	//////////////////////////////////////////////////////////////////////
	// callbacks, required by Expat library
	//////////////////////////////////////////////////////////////////////

    // Called when Expat found a new element (more exactly - its header)
	// Arguments:
	//	data:	actual parser object
	//	el:		name of the found element
	//	attr:	attributes: even items: names, odd: values (text)
    void parser::start_element(void* data, const char* el, const char** attr)
    {
        parser& self = *(parser*)data;
		element *elm = NULL;
		self.m_level ++;

#ifdef TEX_EXTENSIONS
        if(!elm && (strcmp(el, "src") == 0 || strcmp(el, "tex:src") == 0)) 
		{
			elm = new tex::src(self.m_element,self.m_level);
		}

        if(!elm && (strcmp(el, "override") == 0 || strcmp(el, "tex:override") == 0)) 
		{
			elm = new tex::override(self.m_element,self.m_level);
		}
#endif //TEX_EXTENSIONS

        if(!elm && strcmp(el, "svg") == 0) 
		{
			elm = new svg(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "g") == 0) 
		{
			elm = new group(self.m_element,self.m_level);
		} 

		if(!elm && strcmp(el, "defs") == 0) 
		{
			elm = new defs(self.m_element,self.m_level);
		} 

		if(!elm && strcmp(el, "path") == 0) 
		{
			elm = new path(self.m_element,self.m_level);
		} 

		if(!elm && strcmp(el, "text") == 0) 
		{
			elm = new text(self.m_element,self.m_level);
		}


		if(!elm && strcmp(el, "tspan") == 0) 
		{
			elm = new tspan(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "textPath") == 0) 
		{
			elm = new text_path(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "tref") == 0) 
		{
			elm = new tref(self.m_element,self.m_level);
		}


		if(!elm && strcmp(el, "stop") == 0) 
		{
			elm = new stop(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "linearGradient") == 0) 
		{
			elm = new linear_gradient(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "radialGradient") == 0) 
		{
			elm = new radial_gradient(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "pattern") == 0) 
		{
			elm = new pattern(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "use") == 0) 
		{
			elm = new use(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "filter") == 0) 
		{
			elm = new filter(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "feGaussianBlur") == 0) 
		{
			elm = new fe_gaussian_blur(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "clipPath") == 0) 
		{
			elm = new clip_path(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "mask") == 0) 
		{
			elm = new mask(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "polyline") == 0) 
		{
			elm = new polyline(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "polygon") == 0) 
		{
			elm = new polygon(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "line") == 0) 
		{
			elm = new line(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "ellipse") == 0) 
		{
			elm = new ellipse(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "circle") == 0) 
		{
			elm = new ellipse(self.m_element,self.m_level);
		}

		if(!elm && strcmp(el, "rect") == 0) 
		{
			elm = new rect(self.m_element,self.m_level);
		}

        //if(!elm && strcmp(el, "<OTHER_ELEMENT>") == 0) 
		//{
		//	elm = new <aclass>(self.m_element,self.m_level); 
        //} . . .

		if (elm) 
		{
		
			self.parse(*elm,attr); //inside parse elm can receive another "id"
								   //attribute, this time with the name

			//give to EVERY element a chance to store itself,
			//under and empty name. A child class can still handle 
			//this case, overriding the on_attribute() function.
			if (!is_valid(elm->id())) 
			{
				elm->on_attribute(self, "id",""); 
			}

			//Inside on_attribute(self, "id","") the elm should be stored by a call to
			//self.store(). It will update the self.m_element field to the
			//stored element. When nothing like this will happen - we have
			//to delete the newly created item (it is very unusual, but I would like 
			//to be sure, that all heap allocated elements are in the m_elems storage.)
			if (self.m_element != elm) 
			{
				delete elm;
			}
		}

    } //parser::start_element 

    //-------------------------------------------------------------
	//Copy constructor of the reference structure - throws an exception
	//at the attempt to copy it, when it contains an element that has
	//to be deleted. Created for the "safety" reasons - I have made such error,
	//once.
	reference::reference(reference &r) 
	{
		if (r.presisted && elm) 
		{
			char buf[128];
            sprintf(buf, "copy constructor of the svg::reference class: An attempt to copy a stored reference to '%s' element", r.name);
            throw exception(buf);
		}
		elm = NULL; 
		presisted = false;
		strcpy(name, r.name);
	}
    //-------------------------------------------------------------
	//Destructor of the reference structure - I have place it here,
	//to underline that it frees the block created with the new operator
	//in the function parser::start_element, above.
	reference::~reference() 
	{
		if (presisted && elm) 
		{
			delete elm; 
		}
		elm = NULL; 
		name[0]=0; 
		presisted = false;
	}
	//-------------------------------------------------------------
	//helper method: parse attributes of given SVG element 
	//(to be called form inside the element's constructor, from start_element())
	//Arguments:
	//	elm:	SVG graph element, that is the owner of these attributes
	//	attr:	attributes - even elements: names, odd - values (text)
    void parser::parse(element &elm, const char** attr)
	{
		//maybe we will find a font: clear the internal buffers:
		m_font_family[0] = 0;
		m_font_style[0] = 0;
		m_font_weight[0] = 0;
		
		char href[SVG_MAX_NAME]; //a buffer for a reference attribute value
								 //we have to keep it for a while, because 
								 //it HAS to appear AFTER the id attribute
								 //(otherwise the sequence of elements will
								 //be wrong, because *this element can
								 //appear between referenced element and its
								 //child elements!

		href[0] = 0;			 //initially: empty string
        
		int i;
        for(i = 0; attr[i]; i += 2)
        {
            if(strcmp(attr[i], "style") == 0) 
			{
				parse_style(elm, attr[i + 1]);  
			}
			else
			if (strcmp(attr[i], "xlink:href") == 0)
			{
				strncpy(href,attr[i + 1],sizeof(href)-1);
				href[sizeof(href)-1] = 0; //just to be sure
			}
            else
            {
				if (!attribute_accumulated(attr[i],attr[i + 1]))
				{
					elm.on_attribute(*this, attr[i], attr[i + 1]);
				}
            }
        }		
		
		if (href[0])
		{
			//element that uses href has to be registered, anyway:
			if (!is_valid(elm.id())) elm.on_attribute(*this, "id", ""); 
			//and only registered element can have get teh hreference id:								   
			elm.on_attribute(*this, "xlink:href", href);
		}

		//if something is changed with the font shape - add the "font"
		//attribute to this element!
		if (m_font_family[0] || m_font_style[0] || m_font_weight[0])
		{
			char spec[SVG_MAX_ATTRIB_VALUE];
			//fill the font buffer:
			font::specification(spec,sizeof(spec),m_font_family,
								m_font_style,m_font_weight,elm.parent());
			elm.on_attribute(*this,"font",spec);
		}
		

	}
    //-------------------------------------------------------------
    // Called when Expat found a new text inside actual element
	// Arguments:
	//	data:	actual parser object
	//	s:		text found
	//	len:	number of the characters
	void parser::content(void* data, const char* s, int len)
    {
        parser& self = *(parser*)data;
		if (self.m_element && self.m_element->level() == self.m_level)
									self.m_element->on_content(self,s,len);
	}
    //-------------------------------------------------------------
    // Called when Expat found the closing tag of an element
	// Arguments:
	//	data:	actual parser object
	//	el:		name of the actual element
	void parser::end_element(void* data, const char* el)
	{
        parser& self = *(parser*)data;
		if (self.m_element && self.m_element->level() == self.m_level)
		{
			self.m_element->commit(self);
			self.m_element = self.m_element->parent();
		}
		self.m_level--;
	}
    //-------------------------------------------------------------
	///////////////////////////////////////////////////////////////////////
	// helper methods, used in parsing
	///////////////////////////////////////////////////////////////////////
	//Returns an id of element that has given name
	//Arguments:
	//	name:	xml id of the element sought.
	//Beware: it always returns an id, but the requested element can 
	//be not available (retrieve return NULL) until call to initialize()
	//It is not optimized function - it performs linear search, with
	//helper lookup table for the most used elements.
	handle parser::get_id(const char *name)
	{
		const char *ename; //element name
		char buf[SVG_MAX_NAME];
		
		//"eat" the unwanted symbols "url(...)" around the reference
		if(strncmp(name,"url(",4) == 0)
		{	
			size_t size = strlen(name) - 4 - 1; //this -1 to skip closing bracket
			if (size >= sizeof(buf)-1) size = sizeof(buf)-1;
			strncpy(buf,name+4,size);
			buf[size] = 0;
			name = (char *) buf;
		}

		//Name in the xlinks has "#" prefix:
		name = (char *) ((name[0] == '#')? name+1:name);

		//check, if this element has been already requested 
		//and is listed in the lookup table:
		handle i, max = m_lookup.size();  
		for (i = 0; i < max; i++)
		{
			reference &ref = m_elems[m_lookup[i]]; //use reference =>
			//because there is no copy constructor, 
			//it would destroy its pointer, otherwise!

			ename = ref.name;

			//OK, we have the name. Let's compare it:
			if (strcmp(ename,name) == 0) break;

			//OK, then try to use the name 
			//delivered by the contained element.
			//Fonts, for example, can have longer
			//name (specification) than the length of the
			//reference.name buffer!
			if (ref.elm) 
			{
				ename = ref.elm->name(); 
				if (strcmp(ename,name) == 0) break;
			}

		}

		//Let's analyse the result:
		if (i < max) //in the lookup table!
		{
			return m_lookup[i];
		}
		else //not in the lookup table: well, we have to search 
			 //all the elements in the graph:
		{
			max = m_elems.size();
			for (i = 0; i < max; i++)
			{
				ename = m_elems[i].name;
				if (strcmp(ename,name) == 0) break;
			}

			if (i >= max)	//not found! - add an empty reference, and hope that
							//it will be filled soon by the store() method!	
			{
				i = m_elems.size();
				reference ref(name);
				m_elems.add(ref);
			}
			m_lookup.add(i);
			return i;
		}
	}

	//-------------------------------------------------------------
	//Assigns the element to the handle
	//Updates internal state of the element
	//Arguments:
	//	id:		the valid (existing!) handle id
	//	elm:	element, to be assigned
	//	cref:	optional: a cross - reference to the additional "use" placeholder
	void parser::assign(handle id, element *elm, handle cref)
	{ 
		if (m_elems[id].elm) 
		{
			delete (m_elems[id].elm); //just to be sure
		}

		//assign the element pointer to the found/created reference
		m_elems[id].assign(elm); //this function marks the elm for deletion
								 //("delete elm;" upon destruction of the m_elems)

		//these are the "friend" line - see declaration in svg_element.h
		//let the element to know: 
		if (!m_elems[id].elm->m_name) //(some fonts and dasharrays can have name already assigned)
		{
			m_elems[id].elm->m_name = m_elems[id].name; //its name ...
		}

		m_elems[id].elm->m_id = id; //...and the id! 
		if (is_valid(cref)) m_elems[id].elm->m_forward_use = cref; //and the cross-reference 

		if (!is_valid(m_elems[id].elm->m_id_last)) 
		{
			m_elems[id].elm->m_id_last = m_elems[id].elm->m_id;
		}

	}

	//-------------------------------------------------------------
	//Registers element under given id (name) from the xml file 
	//It is common practice to call this function for each regular
	//(i.e. from xml) element twice: once with an empty name ("") -
	//just to reserve the storage id in proper order. 
	//Then, second time, with name. 
	//Arguments:
	//	elm:		element to be registered;
	//	name:		unique name (usually - the xml id attribute) of the elm
	//				If longer than allowed max - a reference with EMPTY name ("")
	//				will be stored. The element should serve its name, instead!
	//	shared:		optional: true, when the element <elm> can be referenced by 
	//				other objects
	//  from_xml:	optional: true, when this is from the xml file, (i.e. there
	//				will be a call to parser::end_element(). When elm is an object,
	//				created just from the scratch, set it to false
	//Updates the fields m_id and initializes m_id_last in the elm object
	void parser::store(element &elm, const char *name, bool shared, bool from_xml)
	{
		handle id;

		if (is_valid(elm.id())) //already stored!
		{
			id = elm.id();
			size_t len = strlen(name);
			if( 0 < len && len < sizeof(m_elems[id].name))//if name fits the size:
			{
				strcpy(m_elems[id].name, name); //just copy the name, only!
			}
		}
		else
		{
			//check, if this element has been already requested 
			//and is listed in the lookup table:
			handle i, max = m_lookup.size(); 
			
			if (name[0] == 0) //when the name is empty - there is nothing to search!
				i = max;
			else
				for (i = 0; i < max; i++)
					if (strcmp(m_elems[m_lookup[i]].name,name) == 0) break;

			//Let's analyse the result:
			if (i < max) //in the lookup table!
			{
				id = m_lookup[i];
				//add a cross-reference
				handle cref = m_elems.size();
				reference ref(""); //just empty ref...
				m_elems.add(ref); //...at the begining: 
				//bind the reference and the element...
				assign(cref,use::create(*this,elm.parent(),elm.level(),id));  
				assign(id,&elm,cref);  //bind the reference and the element...
			}
			else //not in the lookup table: brand new element!
			{
				id = m_elems.size();
				reference ref(name); //just empty ref...
				m_elems.add(ref); //at the begining: just empty ref...
				if (shared)	m_lookup.add(id);
				assign(id,&elm);  //bind the reference and the element...
			}

			
		}

		if (from_xml) m_element = &elm; //element is stored: remember it as the last one!
	}

	
	//-------------------------------------------------------------
	//Stores the attribute in the storage (maintained by the root)
	//Arguments:
	//	elm:	the SVG graph element, that attribute belongs to
	//	attr:	an attribute instance (for example - a local variable)
	//Updates the internal state of the object elm - initializes (if needed)
	//m_attribute_first and updates m_attribute_last fields
	void parser::store(element &elm, attribute attr)
	{
		elm.m_attribute_last = m_attrs.size();
		m_attrs.add(attr);
		if (elm.m_attribute_first == invalid_handle)
		{
			elm.m_attribute_first = elm.m_attribute_last;
		}
	}
	//-------------------------------------------------------------
	//Resolves the painter symbol from a "fill" or "stroke" attribute
	//	- a color
	//	- an url to gradient or pattern
	//Arguments:
	//	elm:	the SVG graph element, that attribute belongs to
	//	type:	fill_flag or stroke_flag
	//	value:	value of an attribute (sometimes - a specification, prepared before)
	//Returns true when the attribute was recognized and stored.
	//Updates the elm intenral stats. (Uses internally the store(elm,attr) 
	//procedure). 
	bool parser::parse_painter(element &elm, attribute::kind type, const char *value)
	{
		//*TODO: add gradient and pattern handling!
        if(strcmp(value, "none") == 0)
        {
			store(elm,attribute(type,false));
			return true;
        }
        else
        {
			char buf[SVG_MAX_ATTRIB_VALUE];


			char *p = get_first_token(value,'#',buf,sizeof(buf));
			if(strcmp(buf,"url(") == 0) //an url of a painter - pattern or gradient
			{	
				//strip the trailing ")" 
				get_first_token(p,')',buf,sizeof(buf)); //buf: element id
				handle id = get_id(buf);	
				if (is_valid(id))
				{
					//OK, we can store this attribute - it points to exeisting object
					store(elm,attribute(type,true));
					store(elm,attribute(attribute::painter(type),id));
					store(elm,attribute(attribute::mode(type),other));
				}
				else //Ignore fills/gradients, that are not previously defined
				{
					store(elm,attribute(type,false));
				}//end if
				return true;
			}
			else //solid color!
			{
				store(elm,attribute(type,true));
				store(elm,attribute(attribute::mode(type),solid));
				store(elm,attribute(attribute::color(type),to_color(value)));
				return true;
			}
		}

		return false;
	}
	//-------------------------------------------------------------
	//Resolves the name and value of an attribute, and stores it in the 
	//storage (maintained by the root)
	//Arguments:
	//	elm:	the SVG graph element, that attribute belongs to
	//	name:	name of an attribute
	//	value:	value of an attribute (sometimes - a specification, prepared before)
	//Returns true when the attribute was recognized as a style and stored.
	//Updates the elm intenral stats. (Uses internally the store(elm,attr) 
	//procedure). It is a kind of "default interpreter of various
	//style attributes.
	bool parser::store_as_style(element &elm, const char * name, const char *value)
	{
		if(strcmp(name, "fill") == 0)
        {
            return parse_painter(elm,attribute::fill_flag,value);
        }
        
        if(strcmp(name, "fill-opacity") == 0)
        {
            store(elm,attribute(attribute::fill_opacity,to_double(value)));
			return true;
        }
        
        if(strcmp(name, "fill-rule") == 0)
        {
            if(strcmp(value, "nonzero") == 0)        
				store(elm,attribute(attribute::fill_rule,agg::fill_non_zero));
            else if(strcmp(value, "evenodd") == 0)  
				store(elm,attribute(attribute::fill_rule,agg::fill_even_odd));
			return true;
        }

        if(strcmp(name, "stroke") == 0)
        {
            return parse_painter(elm,attribute::stroke_flag,value);
        }
        
        if(strcmp(name, "stroke-opacity") == 0)
        {
            store(elm,attribute(attribute::stroke_opacity,to_double(value)));
			return true;
        }
        
        if(strcmp(name, "stroke-width") == 0)
        {
            store(elm,attribute(attribute::stroke_width,to_length(value)));
			return true;
        }
        
        if(strcmp(name, "stroke-linecap") == 0)
        {
            if(strcmp(value, "butt") == 0)        
				store(elm,attribute(attribute::stroke_linecap,agg::butt_cap));
            else if(strcmp(value, "round") == 0)  
				store(elm,attribute(attribute::stroke_linecap,agg::round_cap));
            else if(strcmp(value, "square") == 0) 
				store(elm,attribute(attribute::stroke_linecap,agg::square_cap));
			//skipped value: inherit
			return true;
        }
        
        if(strcmp(name, "stroke-linejoin") == 0)
        {
            if(strcmp(value, "miter") == 0)        
				store(elm,attribute(attribute::stroke_linejoin,agg::miter_join));
            else if(strcmp(value, "round") == 0)  
				store(elm,attribute(attribute::stroke_linejoin,agg::round_join));
            else if(strcmp(value, "bevel") == 0) 
				store(elm,attribute(attribute::stroke_linejoin,agg::bevel_join));
			//skipped value: inherit
			return true;
        }
        
        if(strcmp(name, "stroke-miterlimit") == 0)
        {
            store(elm,attribute(attribute::stroke_miterlimit,to_double(value)));
			return true;
        }
        
        if(strcmp(name, "stroke-dasharray") == 0)
        {
			if (strcmp(value,"none") != 0) //when stroke-dasharray:none ->do nothing
			{
				char spec[sizeof(m_style_value)];
				remove_spaces(value,spec,sizeof(spec)); //just to make it shorter...
				//inside the line below creates and stores a new element - svg::dash
				//and returns its id.
				handle id = get_id(attribute::stroke_dasharray, elm, spec);	
				//OK, we can store this attribute - it points to exeisting object
				store(elm,attribute(attribute::stroke_dasharray,id));
			}
			return true;
        }
        
        if(strcmp(name, "stroke-dashoffset") == 0)
        {
            store(elm,attribute(attribute::stroke_dashoffset,to_double(value)));
			return true;
        }
        
        if(strcmp(name, "font") == 0)
        {
			handle id = get_id(attribute::font, elm, value);	
			//OK, we can store this attribute - it points to existing object
			store(elm,attribute(attribute::font,id));
			return true;
        }
        
        if(strcmp(name, "font-size") == 0)
        {
            store(elm,attribute(attribute::font_size,to_length(value)));
			return true;
        }
        
        if(strcmp(name, "text-anchor") == 0)
        {
            if(strcmp(value, "start") == 0)        
				store(elm,attribute(attribute::text_anchor,(float) 0.0));
            else if(strcmp(value, "middle") == 0)  
				store(elm,attribute(attribute::text_anchor,(float) 0.5));
            else if(strcmp(value, "end") == 0) 
				store(elm,attribute(attribute::text_anchor,(float) 1.0));
			return true;
        }

        if(strcmp(name, "writing-mode") == 0)
        {
            if(strcmp(value, "lr") == 0 || strcmp(value, "lr-tb") == 0)        
				store(elm,attribute(attribute::writing_mode,lr));
            else if(strcmp(value, "rl") == 0 || strcmp(value, "rl-tb") == 0)  
				store(elm,attribute(attribute::writing_mode,rl));
            else if(strcmp(value, "tb") == 0 || strcmp(value, "tb-rl") == 0) 
				store(elm,attribute(attribute::writing_mode,tb));
			return true;
        }

        if(strcmp(name, "kerning") == 0) //dynamic fixes to the distance among letters
        {
            if(strcmp(value, "auto") == 0)        
				store(elm,attribute(attribute::kerning,-1.0)); //negative value = auto.
            else  
	            store(elm,attribute(attribute::kerning,to_length(value)));
			return true;
        }

		//When we are here: the 
		return false;
	}//store_as_style (attribute)
	//--------------------------------------------------------------------
	//helper function, that removes the whitespaces from the source
	//stores the result in another buffer.
	// 
	//Arguments:
	//	srcbuf:		source string;
	//	dstbuf:		destination string
	//	max_size:	maximum number of characters that can be copied into dst
	//				(without trailing null)
	//Returns:  buf, which contains the copied text
	//			NULL, when the src is longer than max_size 
	char *parser::remove_spaces(const char *srcbuf, char *dstbuf, size_t max_size)
	{
		if (strlen(srcbuf) <= max_size)
		{
			//copy the letters, omit the spaces:
			char *src = (char *) srcbuf;
			char *dst = dstbuf; //it is after add prefix
			while (*src)
			{
				if (!isspace(*src)) 
				{
					*dst = *src;
					dst++;
				}
				src++;
			}

			*dst = 0; //close the result.
			return dstbuf;
		}
		else return NULL;
	}
	//--------------------------------------------------------------------
	//helper function that returns a handle to an object, that matches the 
	//requested specification. It tries to find it, and, when it fails, creates 
	//and stores a new instance - as a child element of the given parent.
	//Arguments:
	//	type:	type of attribute, which determines the class of the object.
	//			that has to be created, when existing one was not found
	//	parent: parent object - it will be a parent of the object, that will be
	//			created when the one that matches the requested specification
	//			was not found.
	//	spec:	the text specifiation, describing the object to create
	//			for a dash array: spans of the lines and gaps, separated commas
	//			for a font: type family name, style, weight, separated by commas
	//Returns:  a handler to STORED svg::dash element
	handle parser::get_id(attribute::kind type, element &parent, const char *spec)
	{
		handle id;

		id = get_id(spec); //finds an existing one, or creates an empty place 
						   //when the spec is longer than SVG_MAX_NAME - this object
						   //is created for this particular parent, only!

		if (!m_elems[id].elm) //element not initialized
		{
			element *elm = NULL;

			switch(type)
			{

			case attribute::stroke_dasharray:
				elm = dash::create(*this, &parent,parent.level()+1,spec);
				break;

#ifdef SVG_FONT_WIN32_INCLUDED
			case attribute::font:
				elm = font_win32::create(*this, &parent,parent.level()+1,spec);
				break;
#endif //SVG_FONT_WIN32_INCLUDED
			}

			if (elm)
			{
				assign(id, elm);
				elm->commit(*this);	//dasharray checks here,
									//if it has even spans, (for example)
			}//end if
		}

		return id;
	}

	//--------------------------------------------------------------------
	//////////////////////////////////////////////////////////////////////
	// some useful static functions
	//////////////////////////////////////////////////////////////////////
    double parser::to_double(const char* str)
    {
        while(*str == ' ') ++str;
        return atof(str);
    }

	agg::rgba8 parser::to_color(const char* str)
    {
        while(*str == ' ') ++str;
        return parse_color(str);
    }
	
	//Because in the moment of invoking this function the 
	//base size, used in the relative length, may be unknown,
	//all percent values are converted into simple decimal, negative
	//value - for further processing later on (during request::initialize,
	//or similar).
	//Arguments:
	//	str:		string, which can look like "14.5cm" or "14.5" or "14.5%"
	//	relative:	optional: if not null, set to true, when a percent value
	//				was specified. Otherwise - set to false.
	//				It is up to the caller, to do something with the relative 
	//				value.
    double parser::to_length(const char* str, bool *relative)
    {
		float size; //buffer for the number
		char unit[3]; //optional unit

		if (relative) *relative = false; //default setting

		//Remember - in the constructor of the parser you should 
		//always call setlocale() function, to set in into English!
		//Otherwise in some countries there may be problems with the
		//decimal character ("," instead of ".")
		switch (sscanf(str,"%f%2s",&size,unit))
		{
		case 0:	throw new exception("to_length: Bad SVG syntax - attribute has no value");
		case 1: return size; // no units 
		case 2:	
			{
				double base = 1.0; //unit to pixels 
				if(strcmp(unit, "%")== 0) 
				{
					base = 0.01;
					if (relative) 
					{
						*relative = true; 
					}
					else
					{
						char msg[128];
						sprintf(msg,"to_length: this relative value is not handled: %s.", str);
						throw new exception(msg);
					}
				}
				else 
				if(strcmp(unit, "px")== 0) base = 1.0;//nothing to do
				else 
				if(strcmp(unit, "pt")== 0) base = 1.25;
				else
				if(strcmp(unit, "pc")== 0) base = 15.0;
				else
				if(strcmp(unit, "mm")== 0) base = 3.543307;
				else
				if(strcmp(unit, "cm")== 0) base = 35.43307;
				else
				if(strcmp(unit, "m")== 0) base = 3543.307;
				else
				if(strcmp(unit, "in")== 0) base = 90.0;
				else
				{
					char msg[128];
					sprintf(msg,"to_length: Unknown unit in the length: %s.", unit);
					throw new exception(msg);
				}
				return size * base;
			}
		default: //I cannot imagine that we will reach this case:
				return 0;
		}
		return 0; //I cannot imagine that we will reach this case:
    }

//----------------------------------------------------------------------------
	//////////////////////////////////////////////////////////////////////
	// local helper functions for creating path data in parser::parse_d()
	//////////////////////////////////////////////////////////////////////
	//------------------------------------------------------------------------
	static void move_to(path_storage &paths, double x,  double y, bool rel)// C, c
	{
		if(rel) paths.rel_to_abs(&x, &y); //does not work after closing a subpath!
		paths.move_to(x, y);
	}
	//------------------------------------------------------------------------
	static void line_to(path_storage &paths, double x,  double y, bool rel)// L, l
	{
		if(rel) paths.rel_to_abs(&x, &y);
		paths.line_to(x, y);
	}

	//------------------------------------------------------------------------
	static void hline_to(path_storage &paths, double x, bool rel)		  // H, h
	{
		double x2 = 0.0;
		double y2 = 0.0;
		if(paths.total_vertices())
		{
			paths.vertex(paths.total_vertices() - 1, &x2, &y2);
			if(rel) x += x2;
			paths.line_to(x, y2);
		}
	}

	//------------------------------------------------------------------------
	static void vline_to(path_storage &paths, double y, bool rel)         // V, v
	{
		double x2 = 0.0;
		double y2 = 0.0;
		if(paths.total_vertices())
		{
			paths.vertex(paths.total_vertices() - 1, &x2, &y2);
			if(rel) y += y2;
			paths.line_to(x2, y);
		}
	}

	//------------------------------------------------------------------------
	static void curve_3(path_storage &paths, double x1, double y1,        // Q, q
											double x,  double y, bool rel)
	{
		if(rel) 
		{
			paths.rel_to_abs(&x1, &y1);
			paths.rel_to_abs(&x,  &y);
		}
		paths.curve3(x1, y1, x, y);
	}

	//------------------------------------------------------------------------
	static void curve_3(path_storage &paths, double x, double y, bool rel)// T, t
	{
		if(rel) 
		{
			paths.curve3_rel(x, y);
		} else 
		{
			paths.curve3(x, y);
		}
	}

	//------------------------------------------------------------------------
	static void curve_4(path_storage &paths, double x1, double y1,       // C, c
											double x2, double y2, 
											double x,  double y, bool rel)
	{
		if(rel) 			
		{
			paths.rel_to_abs(&x1, &y1);
			paths.rel_to_abs(&x2, &y2);
			paths.rel_to_abs(&x,  &y);
		}
		paths.curve4(x1, y1, x2, y2, x, y);
	}

	//------------------------------------------------------------------------
	static void curve_4(path_storage &paths, double x2, double y2,      // S, s
											double x,  double y, bool rel)
	{
		if(rel) 
		{
			paths.curve4_rel(x2, y2, x, y);
		} else 
		{
			paths.curve4(x2, y2, x, y);
		}
	}

	//------------------------------------------------------------------------
	static void arc(path_storage &paths, double rx, double ry,		   // A, a
													double angle,
													bool large_arc_flag,
													bool sweep_flag,
													double x, double y, bool rel)
	{
	  if(rel)
	  {
		  paths.arc_rel(rx, ry, angle, large_arc_flag, sweep_flag, x, y);
	  } else
	  {
		  paths.arc_to(rx, ry, angle, large_arc_flag, sweep_flag, x, y);
	  }
	}

	//------------------------------------------------------------------------
	static void close_subpath(path_storage &paths)
	{
		paths.end_poly(agg::path_flags_close);
	}

	//-------------------------------------------------------------
	//parse the "d" attribute - path coordinates and stores it into a path_storage
	//Arguments:
	//	str:	the value of the svg::d xml attribute - geometry of the path		
	//Returns a handle. It is a handle to a path, not to an attribute!
    handle parser::parse_d(const char* str)
	{
		handle id; 
		double prev_x = 0, prev_y = 0; //absolute coordinates of the end
									   //of previous subpath
		m_tok.set_path_str(str);
		id = m_paths.start_new_path();

	    while(m_tok.next())
        {
	        double arg[10];
            char cmd = m_tok.last_command();
            unsigned i;
            switch(cmd)
            {
                case 'M': case 'm':
                    arg[0] = m_tok.last_number();
                    arg[1] = m_tok.next(cmd);
					
					if (cmd == 'm' && !agg::is_vertex(m_paths.vertices().last_command())) //In AGG all "move to" are absolute!
					{
						arg[0] += prev_x;
						arg[1] += prev_y;
						cmd = 'M';
					}
					
					move_to(m_paths, arg[0], arg[1], cmd == 'm');

					//preserve the last "move to" position - it will be useful 
					//when you ecounter next relative "move_to".
					m_paths.vertices().last_vertex(&prev_x, &prev_y);
                    break;

                case 'L': case 'l':
                    arg[0] = m_tok.last_number();
                    arg[1] = m_tok.next(cmd);
                    line_to(m_paths, arg[0], arg[1], cmd == 'l');
                    break;

                case 'V': case 'v':
                    vline_to(m_paths, m_tok.last_number(), cmd == 'v');
                    break;

                case 'H': case 'h':
                    hline_to(m_paths, m_tok.last_number(), cmd == 'h');
                    break;
                
                case 'Q': case 'q':
                    arg[0] = m_tok.last_number();
                    for(i = 1; i < 4; i++)
                    {
                        arg[i] = m_tok.next(cmd);
                    }
                    curve_3(m_paths, arg[0], arg[1], arg[2], arg[3], cmd == 'q');
                    break;

                case 'T': case 't':
                    arg[0] = m_tok.last_number();
                    arg[1] = m_tok.next(cmd);
					curve_3(m_paths, arg[0], arg[1], cmd == 't');
                    break;

                case 'C': case 'c':
                    arg[0] = m_tok.last_number();
                    for(i = 1; i < 6; i++)
                    {
                        arg[i] = m_tok.next(cmd);
                    }
                    curve_4(m_paths, arg[0], arg[1], arg[2], arg[3], arg[4], arg[5], 
						   cmd == 'c');
                    break;

                case 'S': case 's':
                    arg[0] = m_tok.last_number();
                    for(i = 1; i < 4; i++)
                    {
                        arg[i] = m_tok.next(cmd);
                    }
                    curve_4(m_paths, arg[0], arg[1], arg[2], arg[3], cmd == 's');
                    break;

                case 'A': case 'a':
                    arg[0] = m_tok.last_number();
                    for(i = 1; i < 7; ++i)
                    {
                        arg[i] = m_tok.next(cmd);
                    }
                    arc(m_paths, arg[0], arg[1], arg[2], 
							(arg[3]==1.0), (arg[4]==1.0), arg[5], arg[6], cmd=='a');
                    break;

                case 'Z': case 'z':
                    close_subpath(m_paths);

                    break;

                default:
                {
                    char buf[128];
                    sprintf(buf, "parse_d: Invalid Command %c", cmd);
                    throw exception(buf);
                }
            }//switch
		}//while
		return id;
	}//parser::parse_d

	//------------------------------------------------------------------------
	//parse the "points" attribute - polygon/polyline coordinates and stores it into a path_storage
	//Arguments:
	//	str:		the value of the svg::points xml attribute - vertices of the polygon/polyline		
	//	close_flag:	true, if the element is svg:polygon, and false for the svg:polyline		
	//the handle, returned by this function, is a handle to a path, not to an attribute!
    handle parser::parse_points(const char* str, bool close_flag)
	{
        double x = 0.0;
        double y = 0.0;

        handle id = m_paths.start_new_path();
        m_tok.set_path_str(str);
        
		if(!m_tok.next()) throw exception("parse_points: Too few coordinates");
        x = m_tok.last_number();
        if(!m_tok.next()) throw exception("parse_points: Too few coordinates");
        y = m_tok.last_number();
        
		m_paths.move_to(x, y);

        while(m_tok.next())
         {
            x = m_tok.last_number();
            if(!m_tok.next()) throw exception("parse_points: Odd number of coordinates");
            y = m_tok.last_number();
            
			m_paths.line_to(x, y);
         }//while

        if(close_flag) close_subpath(m_paths);
		
        return id;

	}//parse_points
//----------------------------------------------------------------------------
	//Helper function - creates a single line segment 
	//Arguments:
	//	x1,y1:	coordinates of the begining
	//	x2,y2:	coordinates of the end
	//the handle, returned by this function, is a handle to a path, not to an attribute!
    handle parser::create_line(double x1, double y1, double x2, double y2)
    {
        handle id = m_paths.start_new_path();
        m_paths.move_to(x1, y1);
        m_paths.line_to(x2, y2);
		return id;
    }
//----------------------------------------------------------------------------
	//Helper function - creates an ellipse
	//Arguments:
	//	cx,cy:	center of the ellipse
	//	rx,ry:	elipse's radii
	//the handle, returned by this function, is a handle to a path, not to an attribute!
    handle parser::create_ellipse(double cx, double cy, double rx, double ry)
    {
        handle id = m_paths.start_new_path();
        m_paths.move_to(cx+rx, cy);
        m_paths.arc_to(rx, ry, 0, true, true, cx-rx,cy);
        m_paths.arc_to(rx, ry, 0, true, true, cx+rx,cy);
		close_subpath(m_paths);
		return id;
    }
//----------------------------------------------------------------------------
	//Helper function - creates a rectangle (optional rounded corners)
	//Arguments:
	//	x,y:	left, top corner of the rectangle (smallest x, y coordinates)
	//	w,h:	width and height of the rectangle
	//	rx,ry:	corner's radii (can be 0)
	//the handle, returned by this function, is a handle to a path, not to an attribute!
    handle parser::create_rect(double x, double y, double w, double h, double rx, double ry)
    {
		//check the radii:
		if(rx < 0) rx = 0;
		if(ry < 0) ry = 0;
		if(agg::is_equal_eps(rx,0.0,SVG_EPSILON)) rx = ry;
		if(agg::is_equal_eps(ry,0.0,SVG_EPSILON)) ry = rx;
		if (rx >= 0.5 * w) rx = 0.5 * w;
		if (ry >= 0.5 * h) ry = 0.5 * h;

        handle id = m_paths.start_new_path();

		if((rx*ry) > 0.0)
		{
			m_paths.move_to(x + rx, y);
			m_paths.line_to(x + w - rx, y);
			m_paths.arc_to(rx, ry, 0, false, true, x + w, y + ry);
			m_paths.line_to(x + w, y + h - ry);
			m_paths.arc_to(rx, ry, 0, false, true, x + w - rx, y + h);
			m_paths.line_to(x + rx, y + h);
			m_paths.arc_to(rx, ry, 0, false, true, x, y + h - ry);
			m_paths.line_to(x, y + ry);
			m_paths.arc_to(rx, ry, 0, false, true, x + rx, y);
		}
		else
		{
			m_paths.move_to(x, y);
			m_paths.line_to(x + w, y);
			m_paths.line_to(x + w, y + h);
			m_paths.line_to(x, y + h);
			m_paths.line_to(x, y);
		}

		close_subpath(m_paths);
		return id;
    }
//----------------------------------------------------------------------------
	//////////////////////////////////////////////////////////////////////
	// local helper functions for parser::parse_style() function
	//////////////////////////////////////////////////////////////////////
    static void copy_style_token(const char* start, const char* end,
										 char *buf, unsigned bufsize)
    {
        unsigned len = unsigned(end - start);
        if(len > bufsize)
        {
			char msg[128];
            sprintf(msg, "parse_style: Too long element - max. %d characters", bufsize);
            throw exception(msg);
        }
        if(len) memcpy(buf, start, len);
        buf[len] = 0;
    }

    //-------------------------------------------------------------
	//helper function for parsing a single style "name:value" pair
	//Arguments:
	//	elm:		element, for which this style is parsed
	//	nv_start:	the first character of the pair;
	//	nv_end:		the last charcter of the pair;
    void parser::parse_style_element(element &elm, const char* nv_start, 
														const char* nv_end)
    {
        const char* str = nv_start;
        while(str < nv_end && *str != ':') ++str;

        const char* val = str;

        // Right Trim
        while(str > nv_start && 
            (*str == ':' || isspace(*str))) --str;

        ++str;

        copy_style_token(nv_start, str, m_style_name, sizeof(m_style_name));

        while(val < nv_end && (*val == ':' || isspace(*val))) ++val;
        
        copy_style_token(val, nv_end, m_style_value, sizeof(m_style_value));

    }
	//-------------------------------------------------------------

    bool parser::attribute_accumulated(const char* name,const char* value)
	{
		unsigned max = 0;
		char * p = NULL;
		if (strcmp(name,"font-family")==0)
		{
			p = (char *) m_font_family; max = sizeof(m_font_family)-1;
		}
		else
		if (strcmp(name,"font-style")==0)
		{
			p = (char *) m_font_style; max = sizeof(m_font_style)-1;
		}
		else
		if (strcmp(name,"font-weight")==0)
		{
			p = (char *) m_font_weight; max = sizeof(m_font_weight)-1;
		}
		
		if (p)
		{
			strncpy(p,value,max);
			if(strlen(value) < max) max = strlen(value);
			p[max]= 0; //trailing null
			return true;
		}
		else
		{
			return false;
		}

	}//attribut_accumulated

    //-------------------------------------------------------------
	//default handler for parsing the style expression
	//Arguments: 
	//	elm:	element, which contains this style attribute
	//	str:	the value of the xml svg::style attribute
	//It recursively calls the elm.on_attribute() for every style item
	//W.J: it does not use get_token() function, because it becomes
	//from the original AGG example, and I have decided not to touch this
	//working code.
	void parser::parse_style(element &elm, const char* str)
	 {

        while(*str)
        {
            // Left Trim
            while(*str && isspace(*str)) ++str;
            const char* nv_start = str;
            while(*str && *str != ';') ++str;
            const char* nv_end = str;

            // Right Trim
            while(nv_end > nv_start && 
                (*nv_end == ';' || isspace(*nv_end))) --nv_end;
            ++nv_end;

            parse_style_element(elm, nv_start, nv_end); //prepares the values m_style_*

		    if(!attribute_accumulated(m_style_name,m_style_value)) 
			{
				elm.on_attribute(*this, m_style_name, m_style_value);
			}

            if(*str) ++str;
        }
	 }
	//------------------------------------------------------------------------
	//Helper function - for various string tokenization
	//Arguments:
	//	text:		source text;
	//	sep:		separator character
	//	buf:		a buffer for the result (first found token)
	//	bufsize:	size of the buffer (including trailing null)
	//Returns: pointer to a first character after the separator, or 
	//the trailing end of the source text, when no more separators were found
	//Fills the buffer with the token found, stripped of whitespaces.
	//This function can be used in many cases of the string iteration
	char *parser::get_first_token(const char *text, char sep, char *buf, size_t bufsize)
	{
		char *src = (char *) text;
		char *dst = buf;
		char *end = buf + bufsize -1; //the ultimate end of the token 
									  //(the rest will be truncated)
		while(*src && *src != sep)
		{
			if (!isspace(*src) && (dst < end))
			{
				*dst = *src; dst++;
			}
			src++;
		}
		*dst = 0; //insert the trailing null in the result
		if (*src == sep) src++; //result has to point to the next token
		return src;
	}
//----------------------------------------------------------------------------
	//////////////////////////////////////////////////////////////////////
	// local helper functions 
	//////////////////////////////////////////////////////////////////////

    //-------------------------------------------------------------
    static bool is_numeric(char c)
    {
        return strchr("0123456789+-.eE", c) != 0;
    }

    //-------------------------------------------------------------
    static unsigned parse_transform_args(const char* str, 
                                         double* args, 
                                         unsigned max_na, 
                                         unsigned* na)
    {
        *na = 0;
        const char* ptr = str;
        while(*ptr && *ptr != '(') ++ptr;
        if(*ptr == 0)
        {
            throw exception("parse_transform_args: Invalid syntax");
        }
        const char* end = ptr;
        while(*end && *end != ')') ++end;
        if(*end == 0)
        {
            throw exception("parse_transform_args: Invalid syntax");
        }

        while(ptr < end)
        {
            if(is_numeric(*ptr))
            {
                if(*na >= max_na)
                {
                    throw exception("parse_transform_args: Too many arguments");
                }
                args[(*na)++] = atof(ptr);
                while(ptr < end && is_numeric(*ptr)) ++ptr;
            }
            else
            {
                ++ptr;
            }
        }
        return unsigned(end - str);
    }

    //-------------------------------------------------------------
    static transform parse_matrix(const char* str)
    {
        double args[6];
        unsigned na = 0;
        unsigned len = parse_transform_args(str, args, 6, &na);
        if(na != 6)
        {
            throw exception("parse_matrix: Invalid number of arguments");
        }
		return agg::trans_affine(args[0], args[1], args[2], args[3], args[4], args[5]);
    }

    //-------------------------------------------------------------
    static transform parse_translate(const char* str)
    {
        double args[2];
        unsigned na = 0;
        unsigned len = parse_transform_args(str, args, 2, &na);
        if(na == 1) args[1] = 0.0;
        return agg::trans_affine_translation(args[0], args[1]);
    }

    //-------------------------------------------------------------
    static transform parse_rotate(const char* str)
    {
        double args[3];
        unsigned na = 0;
        unsigned len = parse_transform_args(str, args, 3, &na);
        if(na == 1) 
        {
            return agg::trans_affine_rotation(agg::deg2rad(args[0]));
        }
        else if(na == 3)
        {
			transform t = agg::trans_affine_translation(-args[1], -args[2]);
            t *= agg::trans_affine_rotation(agg::deg2rad(args[0]));
            t *= agg::trans_affine_translation(args[1], args[2]);
            return t;
        }
        else
        {
            throw exception("parse_rotate: Invalid number of arguments");
        }
    }

    //-------------------------------------------------------------
    static transform parse_scale(const char* str)
    {
        double args[2];
        unsigned na = 0;
        unsigned len = parse_transform_args(str, args, 2, &na);
        if(na == 1) args[1] = args[0];
        return agg::trans_affine_scaling(args[0], args[1]);
    }

    //-------------------------------------------------------------
    static transform  parse_skew_x(const char* str)
    {
        double arg;
        unsigned na = 0;
        unsigned len = parse_transform_args(str, &arg, 1, &na);
        return agg::trans_affine_skewing(agg::deg2rad(arg), 0.0);
    }

    //-------------------------------------------------------------
    static transform parse_skew_y(const char* str)
    {
        double arg;
        unsigned na = 0;
        unsigned len = parse_transform_args(str, &arg, 1, &na);
        return agg::trans_affine_skewing(0.0, agg::deg2rad(arg));
    }

    //-------------------------------------------------------------
	transform parser::parse_transform(const char* str)
    {
       if(strncmp(str, "matrix", 6) == 0)    return parse_matrix(str);    else 
       if(strncmp(str, "translate", 9) == 0) return parse_translate(str); else 
       if(strncmp(str, "rotate", 6) == 0)    return parse_rotate(str);    else 
       if(strncmp(str, "scale", 5) == 0)     return parse_scale(str);     else 
       if(strncmp(str, "skewX", 5) == 0)     return parse_skew_x(str);    else 
       if(strncmp(str, "skewY", 5) == 0)     return parse_skew_y(str);    else
        {
   			char msg[128];
            sprintf(msg, "parse_transfrom: cannot parse expression '%s'", str);
			throw exception(msg);
        }
    }

} // namespace svg
