//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_TREF_INCLUDED
#include "svg_tspan.h"
#define SVG_TREF_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// tspan: a text content in the SVG graph
//////////////////////////////////////////////////////////////////////
class tref : public tspan
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		tref(element *parent, unsigned level): tspan(parent, level) 
		{ m_href = invalid_handle;}

		virtual const char * tag() {return "tref";};

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (
					(strcmp(name, "xlink:href") == 0) 
			   )
			{	
				m_href  = prs.get_id(value); //id of the path
			}
			else
				tspan::on_attribute(prs,name,value);
		}//on_attribute

		//request handler:
		virtual handle on_request(request &req) 
		{
			switch(req.type())
			{
			case request::initialize:	
				return initialize(req);

			default: 
				return tspan::on_request(req);
			}//switch
		}
		
	public:

	protected:

		//initializes the text content 
		//(copying it from the referenced text)
		handle initialize(request &req)
		{
			text *ref = (text *) req.get_element(m_href,"text");
			if(ref)
			{
				tspan *reft = ref->content();

				if(reft)
				{
					m_text = reft->content();
				}//end if reft
			}//end if ref
			return skip(req);
		}//end initialize

	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		handle	m_href; //handle to the referenced svg::text element
}; //tref

}; //namespace svg
#endif //SVG_TREF_INCLUDED

